/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
///
/// @file cubired0.h
///
/// This file contains the simplest (and inefficient) implementation
/// of procedures for the geometric reduction of full cubical sets,
/// also in the context of cubical maps.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2020 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started in January 2002. Last revision: February 2, 2018.


#ifndef _CHOMP_HOMOLOGY_CUBIRED0_H_
#define _CHOMP_HOMOLOGY_CUBIRED0_H_

#include "chomp/system/config.h"
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/homology/chains.h"
#include "chomp/struct/bitfield.h"
#include "chomp/struct/integer.h"
#include "chomp/struct/hashsets.h"
#include "chomp/struct/setunion.h"
#include "chomp/struct/autoarray.h"
#include "chomp/homology/gcomplex.h"
#include "chomp/cubes/pointbas.h"
#include "chomp/cubes/cube.h"
#include "chomp/cubes/cell.h"
#include "chomp/cubes/cubmaps.h"
#include "chomp/cubes/neighbor.h"
#include "chomp/homology/bddacycl.h"
#include "chomp/homology/tabulate.h"
#include "chomp/homology/known.h"
#include "chomp/homology/cubacycl.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace chomp {
namespace homology {
namespace reduction0 {


// --------------------------------------------------
// ------------------ reduce cubes ------------------
// --------------------------------------------------

/// Reduces a pair of sets of cubes for relative homology computation.
/// The pair is (<i>maincset</i> + <i>cset</i> + <i>other</i>,
/// <i>other</i>), where the cubes in <i>maincset</i> are not to be
/// removed (this is like the "core" of the pair that should be fixed),,
/// the cubes in <i>cset</i> and <i>other</i> are to be considered
/// for removal and could be removed if this is found feasible.
/// The sets <i>maincset + cset</i> and <i>other</i> are disjoint.
/// Does not remove any cubes that are listed in the set <i>keep</i>.
/// Does additional verification, e.g., makes sure
/// that the acyclicity of the given map is preserved.
/// If 'quiet' is set to true then suppresses any messages.
/// Returns the number of cubes removed from both sets.
template <class tCube, class tVerify>
int_t cubreducequiet (const hashedset<tCube> &maincset,
	hashedset<tCube> &cset, hashedset<tCube> &other,
	const hashedset<tCube> &keep, const tVerify &verify,
	bool quiet = true)
{
	// remember the initial number of cubes in both sets
	int_t prev = cset. size () + other. size ();

	// if the case is trivial, return the answer
	if (maincset. empty () && cset. empty ())
	{
		if (!other. empty ())
		{
			hashedset<tCube> empty;
			other = empty;
		}
		return prev;
	}

	// determine the space dimension
	int dim = (cset. empty ()) ? maincset [0]. dim () :
		cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int_t maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield for storing neighborhood configurations
	// and allocate the buffer if necessary (this is a static variable)
	static BitField b;
	static int_t _maxneighbors = 0;
	static auto_array<unsigned char> buffer;
	if (maxneighbors != _maxneighbors)
	{
		_maxneighbors = maxneighbors;
		buffer. reset (new unsigned char [(maxneighbors + 7) >> 3]);
		b. define (buffer. get (), maxneighbors);
	}

	// prepare a counter for displaying the progress of computations
	int_t count = 0;

	// repeat scanning cubes until no change can be made
	while (1)
	{
		// remember if any changes were made in this round
		bool modified (false);

		// scan all the cubs in both sets
		for (int_t i = 0; i < cset. size () + other. size (); ++ i)
		{
			// show the progress indicator
			++ count;
			if (!quiet && !(count % 373))
				scon << std::setw (10) << count <<
					"\b\b\b\b\b\b\b\b\b\b";

			// is this cube from the other set?
			bool is_other (i >= cset. size ());

			// determine the cube corresponding to the index
			const tCube &c (is_other ?
				other [i - cset. size ()] : cset [i]);

			// if this cube is to be kept then skip it
			if (keep. check (c))
				continue;

			// clean the neighborbits
			b. clearall (maxneighbors);

			// if this cube belongs to the other set...
			if (is_other)
			{
				// no neighbors need to be remembered
				hashedset<tCube> *no_neighbors (0);

				// check if this cube can be removed
				if (!acyclic_rel (c, dim,
					makesetunion (maincset, cset),
					other, &b, maxneighbors,
					no_neighbors, no_neighbors))
				{
					continue;
				}

				// verify if the removal of the cube from A is OK
				if (!verify (c, other))
					continue;

				// verify if the removal of the cube from X is OK
				if (!verify (c, makesetunion (maincset,
					makesetunion (other, cset))))
				{
					continue;
				}

				// remove the cube from the other set
				modified = true;
				if (!quiet)
					sseq << '0' << c << '\n';
				other. removenum (i - cset. size ());
				-- i;
			}

			// otherwise, if this cube belongs to 'cset'...
			else
			{
				// no neighbors need to be remembered
				hashedset<tCube> *no_neighbors (0);

				// if the neighborhood of the cube
				// is not acyclic then skip the cube
				if (!acyclic (c, dim, makesetunion (maincset,
					makesetunion (cset, other)),
					&b, maxneighbors, no_neighbors))
				{
					continue;
				}

				// verify if the removal of the cube from X is OK
				if (!verify (c, makesetunion (maincset,
					makesetunion (cset, other))))
				{
					continue;
				}

				// remove the cube from 'cset'
				modified = true;
				if (!quiet)
					sseq << '0' << c << '\n';
				cset. removenum (i);
				-- i;
			}
		}

		// if the sets were not modified then quit the loop
		if (!modified)
			break;
	}

	// return the number of cubes removed from both sets
	return prev - cset. size () - other. size ();
} /* cubreducequiet */


// --------------------------------------------------
// ------------------ expand cubes ------------------
// --------------------------------------------------

/// Expands the set 'other' towards 'cset' without changing the homology
/// of (cset + other, other). The two sets must be disjoint.
/// Use an externally defined additional criterion
/// for the verification if the cubes that look good for removal
/// are actually allowed to be removed.
template <class tCube, class tVerify>
int_t cubexpand (hashedset<tCube> &cset, hashedset<tCube> &other,
	const tVerify &verify, bool quiet = false)
{
	// if the case is trivial, return the answer
	if (cset. empty () || other. empty ())
		return 0;

	// remember the initial number of cubes in the main set
	int_t prev = cset. size ();

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int_t maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int_t _maxneighbors = 0;
	static auto_array<unsigned char> buffer;
	if (maxneighbors != _maxneighbors)
	{
		_maxneighbors = maxneighbors;
		buffer. reset (new unsigned char [(maxneighbors + 7) >> 3]);
		b. define (buffer. get (), maxneighbors);
	}

	// prepare a counter for displaying the progress of computations
	int_t count = 0;

	// repeat scanning cubes until no change can be made
	while (1)
	{
		// remember if any changes were made in this round
		bool modified (false);

		// go through the main set, and move cubes to the other set
		// if this doesn't change the homology
		for (int_t i = 0; i < cset. size (); ++ i)
		{
			// show progress indicator
			++ count;
			if (!quiet && !(count % 373))
				scon << std::setw (10) << count <<
					"\b\b\b\b\b\b\b\b\b\b";

			// take a cube from 'cset'
			const tCube &c = cset [i];

			// clear the neighborbits
			b. clearall (maxneighbors);

			// if the neighborhood of the cube is not acyclic, skip it
			if (!acyclic (c, dim, other, &b, maxneighbors))
				continue;

			// verify if moving this cube to 'other' is allowed
			if (!verify (c, cset, other))
				continue;

			// move the cube to the other set
			modified = true;
			if (!quiet)
				sseq << "L\n" << '2' << c << '\n';
			other. add (c);
			cset. removenum (i);
			-- i;
		}

		// if the sets were not modified then quit the loop
		if (!modified)
			break;
	}

	// return the number of cubes removed from 'cset'
	return prev - cset. size ();
} /* cubexpand */


} // namespace reduction0
} // namespace homology
} // namespace chomp

#endif // _CHOMP_HOMOLOGY_CUBIRED0_H_

/// @}

